package pa.ub.sorting;

import java.nio.IntBuffer;
import java.util.Random;

import org.lwjgl.BufferUtils;
import org.lwjgl.PointerBuffer;
import org.lwjgl.opencl.CL10;
import org.lwjgl.opencl.CLCommandQueue;
import org.lwjgl.opencl.CLContext;
import org.lwjgl.opencl.CLKernel;
import org.lwjgl.opencl.CLMem;
import org.lwjgl.opencl.CLProgram;

import pa.cl.CLUtil;
import pa.cl.CLUtil.PlatformDevicePair;
import pa.cl.OpenCL;
import pa.util.IOUtil;
import pa.util.SizeOf;

public class BucketSortNaivSlow {
    
    private static CLContext context;
    
    private static CLKernel fillbuckets, evaluate;
    
    private static CLCommandQueue queue;
    private static CLProgram program;
    private static CLKernel reduceForward;
    private static CLKernel reduceBackward;
    
    private static int count = 1 << 22;
    
    private static final int N = 256; // Alphabetgroesse N
    
    private static PointerBuffer gws = new PointerBuffer(1);
    
    //cpu version http://www-lehre.inf.uos.de/~ainf/2012/PDF/skript.pdf
    public static void sort (char[] a) { // sortiere Character-Array a
        
        int[] b = new int[N]; // N Buckets
        int i, j, k; // Laufvariablen

        for (i = 0; i < N; i++)
            b[i] = 0; // setze alle Buckets auf 0
        
        for (i = 0; i < a.length; i++)
         // fuer jedes Eingabezeichen
            b[a[i]]++; // zustaendiges Bucket erhoehen
            k = 0;
            for (i = 0; i < N; i++)
                // fuer jedes Bucket
                for (j = 0; j < b[i]; j++)
                    // gemaess Zaehlerstand
                    a[k++] = (char) i; // sein Zeichen uebernehmen
    }
    
    public static void main(String[] args)
    {        
        CLUtil.createCL();
        
        PlatformDevicePair pair = CLUtil.choosePlatformAndDevice();
        
        IntBuffer glyphs = BufferUtils.createIntBuffer(count);

        Random r = new Random(0);
        char[] chars = new char[count];
        for(int i = 0; i < count; ++i)
        {
            int c = r.nextInt(256);
            glyphs.put(i, c);
            chars[i] = (char)c;
        }
        
        context = OpenCL.clCreateContext(pair.platform, pair.device, null, null);
        queue = OpenCL.clCreateCommandQueue(context, pair.device, OpenCL.CL_QUEUE_PROFILING_ENABLE);
        program = OpenCL.clCreateProgramWithSource(context, IOUtil.readFileContent("kernel/sort.cl"));
        OpenCL.clBuildProgram(program, pair.device, "", null);
        
        fillbuckets = OpenCL.clCreateKernel(program, "fillBucket");
        evaluate = OpenCL.clCreateKernel(program, "eval");
        reduceForward = OpenCL.clCreateKernel(program, "reduce");
        reduceBackward = OpenCL.clCreateKernel(program, "reduceInverse");
        
        IntBuffer cb = BufferUtils.createIntBuffer(N);
        for(int i = 0; i < N; ++i)
        {
            cb.put(i, 0);
        }

        CLMem clBuckets = OpenCL.clCreateBuffer(context, OpenCL.CL_MEM_READ_WRITE | OpenCL.CL_MEM_COPY_HOST_PTR, cb);
        CLMem clGlphys = OpenCL.clCreateBuffer(context, OpenCL.CL_MEM_READ_WRITE | OpenCL.CL_MEM_COPY_HOST_PTR, glyphs);
        CLMem clScan = OpenCL.clCreateBuffer(context, OpenCL.CL_MEM_READ_WRITE | OpenCL.CL_MEM_COPY_HOST_PTR, cb);

        OpenCL.clSetKernelArg(fillbuckets, 0, clBuckets);
        OpenCL.clSetKernelArg(fillbuckets, 1, clGlphys);
        
        OpenCL.clSetKernelArg(evaluate, 0, clScan);
        OpenCL.clSetKernelArg(evaluate, 1, clBuckets);
        OpenCL.clSetKernelArg(evaluate, 2, clGlphys);
        
        OpenCL.clSetKernelArg(reduceForward, 0, clScan);
        OpenCL.clSetKernelArg(reduceBackward, 0, clScan);
        
        gws.put(0, count);
        
        long t = CLUtil.measureKernelCall(fillbuckets, queue, gws, null, 1);
        
        CL10.clEnqueueCopyBuffer(queue, clBuckets, clScan, 0, 0, cb.capacity() * SizeOf.INTEGER, null, null);
        
        gws.put(0, N / 2);
        
        t += CLUtil.measureKernelCall(reduceForward, queue, gws, gws, 1);
        
        OpenCL.clEnqueueReadBuffer(queue, clScan, 1, 0, cb, null, null);
        
        cb.put(0, 0);
        
        OpenCL.clEnqueueWriteBuffer(queue, clScan, 1, 0, cb, null, null);
        
        t += CLUtil.measureKernelCall(reduceBackward, queue, gws, gws, 1);
        
        gws.put(0, N);
        
        t += CLUtil.measureKernelCall(evaluate, queue, gws, gws, 1);
        
        /*OpenCL.clEnqueueReadBuffer(queue, clBuckets, 1, 0, cb, null, null);
        
        BufferHelper.printBuffer(cb, N);
        
        OpenCL.clEnqueueReadBuffer(queue, clScan, 1, 0, cb, null, null);
        
        BufferHelper.printBuffer(cb, N); */
        
        System.out.println("BucketSort: " + count + " Elements");
        
        long start = System.nanoTime();
        sort(chars);
        long end = System.nanoTime();
        /*for(int i = 0; i < chars.length; ++i)
        {
            text += (int)chars[i] + (i < chars.length - 1 ? ", " : "");
        } */
        System.out.println("CPU Time: " + (end - start) / 1e9f + " Seconds.");
        System.out.println("GPU Time: " + t / 1e9f + " Seconds.");
        
        
        /*System.out.print("Unsorted  : ");
        for(int i = 0; i < glyphs.capacity(); ++i)
        {
            System.out.print(glyphs.get(glyphs.capacity() - i - 1) + ((i == glyphs.capacity()-1) ? "" : ", "));
        }
        System.out.println();
        
        OpenCL.clEnqueueReadBuffer(queue, clGlphys, 1, 0, glyphs, null, null);

        System.out.print  ("Sorted    : ");
        for(int i = 0; i < glyphs.capacity(); ++i)
        {
            System.out.print(glyphs.get(glyphs.capacity() - i - 1) + ((i == glyphs.capacity()-1) ? "" : ", "));
        } */

        
        CLUtil.destroyCL();
    }
}
