package pa.ub.sorting;

import java.nio.IntBuffer;
import java.util.Random;

import org.lwjgl.BufferUtils;
import org.lwjgl.PointerBuffer;
import org.lwjgl.opencl.CL10;
import org.lwjgl.opencl.CLCommandQueue;
import org.lwjgl.opencl.CLContext;
import org.lwjgl.opencl.CLKernel;
import org.lwjgl.opencl.CLMem;
import org.lwjgl.opencl.CLProgram;

import pa.cl.CLUtil;
import pa.cl.CLUtil.PlatformDevicePair;
import pa.cl.OpenCL;
import pa.util.BufferHelper;
import pa.util.IOUtil;
import pa.util.SizeOf;

public class RadixSort {
    
    private static CLContext context;
    private static CLKernel sort;
    
    private static CLKernel reduceForward;
    private static CLKernel reduceBackward;
    private static CLKernel compact;
    
    private static CLCommandQueue queue;
    private static CLProgram program;
    
    private static CLMem clNumbers, clParts, clBools, clScan;
    
    private static int elementCount = 8;
    
    private static IntBuffer numbers = BufferUtils.createIntBuffer(elementCount);
    
    private static IntBuffer parts = BufferUtils.createIntBuffer(2 * elementCount);
    
    private static IntBuffer partsTmp = BufferUtils.createIntBuffer(2 * elementCount);
    
    private static IntBuffer bools;
    
    private static IntBuffer tmp;
    
    private static Random r = new Random(0);
    
    private static PointerBuffer gws = new PointerBuffer(1);
      
    private static int MAX_VAL = 803;
    
    private static int digits = getPossibleBitsToUse(MAX_VAL+1);
    
    private static int getPossibleBitsToUse(int max)
    {
        int max2pow = Integer.highestOneBit(MAX_VAL);
        int pow = 1;
        int i = 0;
        while(pow <= max2pow)
        {
            pow *= 2;
            ++i;
        }
        return i;
    }
    
    public static void sortDigit()
    {
        gws.put(0, elementCount);
        
        OpenCL.clEnqueueWriteBuffer(queue, clBools, 0, 0, bools, null, null);
        OpenCL.clEnqueueWriteBuffer(queue, clParts, 0, 0, parts, null, null);
        
        OpenCL.clEnqueueNDRangeKernel(queue, sort, 1, null, gws, null, null, null);
        
        OpenCL.clEnqueueWriteBuffer(queue, clScan, 0, 0, bools, null, null);

        gws.put(0, bools.capacity() / 2);
        
        CL10.clEnqueueCopyBuffer(queue, clBools, clScan, 0, 0, bools.capacity() * SizeOf.INTEGER, null, null);
        
        OpenCL.clEnqueueNDRangeKernel(queue, reduceForward, 1, null, gws, gws, null, null);
        
        /*OpenCL.clEnqueueReadBuffer(queue, clScan, 1, 0, tmp, null, null);
        
        BufferHelper.printBuffer(tmp, elementCount, "half"); */
        int max = tmp.get(0);
        tmp.put(0, 0);
        
        OpenCL.clEnqueueWriteBuffer(queue, clScan, 1, 0, tmp, null, null);
        
        OpenCL.clEnqueueNDRangeKernel(queue, reduceBackward, 1, null, gws, gws, null, null);
        
        gws.put(0, 2 * elementCount);
        OpenCL.clSetKernelArg(compact, 5, max);
        OpenCL.clEnqueueNDRangeKernel(queue, compact, 1, null, gws, gws, null, null);
        /*
        OpenCL.clEnqueueReadBuffer(queue, clBools, 1, 0, tmp, null, null);
        
        BufferHelper.printBuffer(tmp, elementCount, "bools");
        
        OpenCL.clEnqueueReadBuffer(queue, clScan, 1, 0, tmp, null, null);
        
        BufferHelper.printBuffer(tmp, elementCount, "scan");
        
        OpenCL.clEnqueueReadBuffer(queue, clParts, 1, 0, partsTmp, null, null);
        
        BufferHelper.printBuffer(partsTmp, elementCount, "parts");
        
        OpenCL.clEnqueueReadBuffer(queue, clNumbers, 1, 0, numbers, null, null);
                
        BufferHelper.printBuffer(numbers, elementCount, "Compact");
        
        System.out.println("----"); */
    }
    
    public static void main(String[] args)
    {
        CLUtil.createCL();

        PlatformDevicePair pair = CLUtil.choosePlatformAndDevice();

        context = OpenCL.clCreateContext(pair.platform, pair.device, null, null);
        queue = OpenCL.clCreateCommandQueue(context, pair.device, OpenCL.CL_QUEUE_PROFILING_ENABLE);
        program = OpenCL.clCreateProgramWithSource(context, IOUtil.readFileContent("kernel/sort.cl"));
        OpenCL.clBuildProgram(program, pair.device, "", null);
        
        sort = OpenCL.clCreateKernel(program, "histo");
        reduceForward = OpenCL.clCreateKernel(program, "reduce");
        reduceBackward = OpenCL.clCreateKernel(program, "reduceInverse");
        compact = OpenCL.clCreateKernel(program, "compact");
        
        clParts = OpenCL.clCreateBuffer(context, OpenCL.CL_MEM_READ_WRITE | OpenCL.CL_MEM_COPY_HOST_PTR, parts);

        int size = 1;
        while(size < elementCount * digits)
        {
            size *= 2;
        }

        bools = BufferUtils.createIntBuffer(2 * elementCount);
        tmp = BufferUtils.createIntBuffer(2 * elementCount);

        for(int i = 0; i < elementCount; ++i)
        {
            numbers.put(i, r.nextInt(MAX_VAL + 1));
        }
        
        for(int i = 0; i < bools.capacity(); ++i)
        {
            bools.put(i, 0);
        }
        
        for(int i = 0; i < parts.capacity(); ++i)
        {
            parts.put(i, 0);
        }
        
        clNumbers = OpenCL.clCreateBuffer(context, OpenCL.CL_MEM_READ_WRITE | OpenCL.CL_MEM_COPY_HOST_PTR, numbers);
        clBools = OpenCL.clCreateBuffer(context, OpenCL.CL_MEM_READ_WRITE | OpenCL.CL_MEM_COPY_HOST_PTR, bools);
        clScan = OpenCL.clCreateBuffer(context, OpenCL.CL_MEM_READ_WRITE | OpenCL.CL_MEM_COPY_HOST_PTR, bools);
        
        OpenCL.clSetKernelArg(reduceForward, 0, clScan);
        OpenCL.clSetKernelArg(reduceBackward, 0, clScan);
        
        OpenCL.clSetKernelArg(sort, 0, clNumbers);
        OpenCL.clSetKernelArg(sort, 1, clParts);
        OpenCL.clSetKernelArg(sort, 2, clBools);
        OpenCL.clSetKernelArg(sort, 3, elementCount);
        
        OpenCL.clSetKernelArg(compact, 0, clScan);
        OpenCL.clSetKernelArg(compact, 1, clNumbers);
        OpenCL.clSetKernelArg(compact, 2, clBools);
        OpenCL.clSetKernelArg(compact, 3, clParts);
        OpenCL.clSetKernelArg(compact, 4, elementCount);
        System.out.println("RadixSort:");
        BufferHelper.printBuffer(numbers, elementCount, "Unsorted: ");

        for(int i = 0; i < digits; ++i)
        {
            OpenCL.clSetKernelArg(sort, 4, i);
            sortDigit();
            //digit *= 10;
        }
        
        OpenCL.clEnqueueReadBuffer(queue, clNumbers, 1, 0, numbers, null, null);
        
        BufferHelper.printBuffer(numbers, elementCount, "Sorted: ");
        
        OpenCL.clReleaseMemObject(clBools);
        OpenCL.clReleaseMemObject(clNumbers);
        OpenCL.clReleaseMemObject(clParts);
        OpenCL.clReleaseMemObject(clScan);
        
        OpenCL.clReleaseKernel(compact);
        OpenCL.clReleaseKernel(reduceBackward);
        OpenCL.clReleaseKernel(reduceForward);
        OpenCL.clReleaseKernel(sort);
        
        OpenCL.clReleaseProgram(program);
        
        OpenCL.clReleaseCommandQueue(queue);
        
        OpenCL.clReleaseContext(context);
        

        CLUtil.destroyCL();
    }
}
