
//--radix sort

kernel void histo(global int* data, global int* part, global int* bools, const int N, const int digit)
{
    uint id = get_global_id(0);
    
    int val = data[id];
    
    int bit = val & (1 << digit); //val / digit % 10;
    int shift = bit == (1 << digit);
    part[shift * N + id] = val;
    bools[shift * N + id] = 1;
}

kernel void reduce(global int* data) 
{
    uint id = get_global_id(0);
    uint blockSize = get_local_size(0);
    uint block = get_group_id(0);
    for(uint s = 1; s <= blockSize; s *= 2)
    {
        if(id <= blockSize-s)
        {
            data[2 * id * s] = data[2 * id * s] + data[s * 2 * id + s];
        }
        barrier(CLK_GLOBAL_MEM_FENCE);
    }
}

kernel void reduceInverse(global int* data) 
{
    uint id = get_global_id(0);
    uint blockSize = get_local_size(0);
    uint stride = 1;
    for(uint s = blockSize; s > 0; s >>= 1)
    {
        if(id < stride)
        {
            int il = 2 * id * s;
            int ir = il + s;
            
            int left = data[il];
            int right = data[ir];
                                
            data[ir] = left;
            
            data[il] = left + right;
        }
        stride *= 2;
        barrier(CLK_GLOBAL_MEM_FENCE);
    }
}

kernel void compact(global int* prefixSum, global int* result, global int* bools, global int* data, const int N, const int max)
{
    uint id = get_global_id(0);
    int exists = bools[id];
    if(exists)
    {
        result[max - prefixSum[id] - 1] = data[id];
    }
}

//bucket sort

kernel void fillBucket(global int* buckets, global uint* glyphs)
{
    uint id = get_global_id(0);
    uint g = glyphs[id];
    atomic_inc(buckets + g);
}

kernel void eval(global int* scan, global int* buckets, global uint* glyphs, uint N)
{   
    uint index = get_global_id(0);
    uint c = buckets[index];
    if(c)
    {
        glyphs[scan[index]] = (index / N);
    }
}